#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
#import "NSString+finder.h"
#include <sys/types.h>

static const char *argv0;
static NSWorkspace *workspace;

static int reveal_cmd(NSString *cmd, NSArray *args);
static int update_cmd(NSString *cmd, NSArray *args);
static int open_cmd(NSString *cmd, NSArray *args);
static int start_cmd(NSString *cmd, NSArray *args);
static int stop_cmd(NSString *cmd, NSArray *args);
static int version_cmd(NSString *cmd, NSArray *args);
static int help_cmd(NSString *cmd, NSArray *args);

static void printVersion(void);
static void printUsage(void);

static const char version[] = "1.0";

#pragma mark Main program

int main(int argc, const char **argv) {
	argv0 = argc ? argv[0] : "finder";

	NSAutoreleasePool *pool = [NSAutoreleasePool new];

	int result = 0;
	if(argc < 2) {
		printUsage();
	} else {
		workspace = [NSWorkspace sharedWorkspace];

		--argc;
		NSString *cmd = [NSString stringWithUTF8String:*++argv];
		NSArray *args = [NSString stringsWithUTF8Strings:++argv count:--argc];

		enum { numFuncs = 11 };
		void *keys[numFuncs]   = { @"reveal",  @"update",   @"updated",  @"open",  @"launch", @"start",  @"stop", @"help", @"--help", @"version", @"--version", };
		void *values[numFuncs] = { reveal_cmd, update_cmd, update_cmd, open_cmd, open_cmd, start_cmd, stop_cmd, help_cmd, help_cmd, version_cmd, version_cmd, };
		CFDictionaryRef cmds = CFDictionaryCreate(kCFAllocatorDefault,
			(const void **)keys, (const void **)values,
			numFuncs,
			&kCFTypeDictionaryKeyCallBacks,
			/*valueCallbacks*/ NULL);
		int (*func)(NSString *,NSArray *) = (int (*)(NSString *,NSArray *))CFDictionaryGetValue(cmds, [cmd lowercaseString]);
		CFRelease(cmds);

		result = func(cmd, args);
	}

	[pool release];
	return -result;
}

static void printVersion(void) {
	printf("finder version %s by Peter Hosey\n", version);
}

static void printUsage(void) {
	printf("finder version %s by Peter Hosey\n"
		   "Control the Finder from the command line.\n"
		   "\n"
		   "Usage:\n"
		   "\t%s update FILE: Tell the Finder to update its information about FILE in any relevant Finder windows.\n"
		   "\t%s reveal FILE: Show FILE in a new Finder window.\n"
		   "\t%s open FILE: Open FILE as if it were double-clicked.\n"
		   "\t%s start: Launch the Finder, if it isn't already running.\n"
		   "\t%s stop: Quit the Finder.\n",
		   version,
		   argv0, argv0, argv0, argv0, argv0);
}

#pragma mark Commands

static int reveal_cmd(NSString *cmd, NSArray *args) {
	NSEnumerator *fileEnum = [args objectEnumerator];
	NSString *path;
	while((path = [fileEnum nextObject]))
		[workspace selectFile:[path absolutePath] inFileViewerRootedAtPath:@""];
	return 0;
}
static int update_cmd(NSString *cmd, NSArray *args) {
	if(![args count])
		[workspace noteFileSystemChanged];
	else {
		NSEnumerator *fileEnum = [args objectEnumerator];
		NSString *path;
		while((path = [fileEnum nextObject]))
			[workspace noteFileSystemChanged:path];
	}
	return 0;
}
static int open_cmd(NSString *cmd, NSArray *args) {
	BOOL anyFailed = NO;
	NSEnumerator *fileEnum = [args objectEnumerator];
	NSString *path;
	while((path = [fileEnum nextObject])) {
		BOOL failed = [workspace openFile:path];
		if(failed) {
			fprintf(stderr, "%s: could not open file %s\n", argv0, [path UTF8String]);
			anyFailed = YES;
		}
	}
	return -anyFailed;
}
static int start_cmd(NSString *cmd, NSArray *args) {
	return ![workspace launchApplication:@"Finder"];
}
static int stop_cmd(NSString *cmd, NSArray *args) {
	NSData *bundleIDData = [@"com.apple.finder" dataUsingEncoding:NSUTF8StringEncoding];
	NSAppleEventDescriptor *target = [NSAppleEventDescriptor descriptorWithDescriptorType:typeApplicationBundleID
																					 data:bundleIDData];
	//create a Quit event.
	NSAppleEventDescriptor *event  = [NSAppleEventDescriptor appleEventWithEventClass:kCoreEventClass
																			  eventID:kAEQuitApplication
																	 targetDescriptor:target
																			 returnID:kAutoGenerateReturnID
																		transactionID:kAnyTransactionID];
	//send it to the Finder.
	OSStatus err = AESend([event aeDesc],
						  /*reply*/ NULL,
						  kAENoReply | kAENeverInteract,
						  kAENormalPriority,
						  /*timeOutInTicks*/ 3600 /*1 minute*/,
						  /*idleProc*/ NULL,
						  /*filterProc*/ NULL);
	if(err != noErr) {
		NSLog(@"AESend returned %li when trying to quit the Finder\n", (long)err);
		return -1;
	} else
		return 0;
}

static int version_cmd(NSString *cmd, NSArray *args) {
	printVersion();
	return 0;
}

static int help_cmd(NSString *cmd, NSArray *args) {
	printUsage();
	return 0;
}
